use axum::{
    extract::{Path, State},
    Json,
};
use axum_extra::routing::Resource;
use http::StatusCode;
use serde::{Deserialize, Serialize};
use sqlx::query;
use ulid::Ulid;
use url::Url;
use uuid::Uuid;

use crate::{app::SharedState, error::AppError};

pub fn resource() -> Resource<SharedState> {
    Resource::named("links")
        .create(create_link)
        .show(get_link_info)
        .update(update_link)
        .destroy(delete_link)
}

#[derive(Serialize)]
struct Link {
    id: Ulid,
    slug: String,
    destination: String,
}

async fn get_link_info(
    State(SharedState { db, .. }): State<SharedState>,
    Path(id): Path<Ulid>,
) -> Result<Json<Link>, AppError> {
    match query!(
        "SELECT id, slug, destination FROM link WHERE id = $1",
        Uuid::from(id),
    )
    .fetch_optional(&db)
    .await?
    {
        Some(r) => Ok(Json(Link {
            id: Ulid::from(r.id),
            slug: r.slug,
            destination: r.destination,
        })),
        None => Err(AppError::LinkNotFoundId(id)),
    }
}

#[derive(Deserialize)]
struct CreateLinkRequestBody {
    slug: String,
    destination: Url,
}

async fn create_link(
    State(SharedState { db, .. }): State<SharedState>,
    Json(CreateLinkRequestBody { slug, destination }): Json<CreateLinkRequestBody>,
) -> Result<Json<Link>, AppError> {
    let id = Ulid::new();

    match query!(
        "INSERT INTO link (id, slug, destination) VALUES ($1, $2, $3) ON CONFLICT DO NOTHING",
        Uuid::from(id),
        slug,
        destination.to_string(),
    )
    .execute(&db)
    .await?
    .rows_affected()
    {
        1 => Ok(Json(Link {
            id,
            slug,
            destination: destination.to_string(),
        })),
        0 => Err(AppError::LinkExists(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}

#[derive(Deserialize)]
struct UpdateLinkRequestBody {
    destination: Url,
}

async fn update_link(
    State(SharedState { db, .. }): State<SharedState>,
    Path(id): Path<Ulid>,
    Json(UpdateLinkRequestBody { destination }): Json<UpdateLinkRequestBody>,
) -> Result<StatusCode, AppError> {
    match query!(
        "UPDATE link SET destination = $2 WHERE id = $1",
        Uuid::from(id),
        destination.to_string(),
    )
    .execute(&db)
    .await?
    .rows_affected()
    {
        1 => Ok(StatusCode::NO_CONTENT),
        0 => Err(AppError::LinkNotFoundId(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}

async fn delete_link(
    State(SharedState { db, .. }): State<SharedState>,
    Path(id): Path<Ulid>,
) -> Result<StatusCode, AppError> {
    match query!("DELETE FROM link WHERE id = $1", Uuid::from(id))
        .execute(&db)
        .await?
        .rows_affected()
    {
        1 => Ok(StatusCode::NO_CONTENT),
        0 => Err(AppError::LinkNotFoundId(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}
