use axum::{body::Body, response::IntoResponse};
use http::StatusCode;
use tracing::{error, field};
use ulid::Ulid;

#[derive(Debug, thiserror::Error)]
pub enum AppError {
    #[error("link already exists ({0})")]
    LinkExists(Ulid),
    #[error("link not found ({0})")]
    LinkNotFoundId(Ulid),
    #[error("link not found ({0})")]
    LinkNotFoundSlug(String),
    #[error("paste not found ({0})")]
    PasteNotFound(Ulid),
    #[error("database returned an impossible number of affected rows ({0})")]
    ImpossibleAffectedRows(u64),
    #[error("ulid conflict")]
    UlidConflict(Ulid),
    #[error("database error: {0}")]
    Database(#[from] sqlx::Error),
    #[error(transparent)]
    Io(#[from] std::io::Error),
    #[error(transparent)]
    Other(#[from] eyre::Report),
}

impl IntoResponse for AppError {
    fn into_response(self) -> axum::http::Response<Body> {
        error!(err = field::display(&self));
        match self {
            Self::LinkExists(_) => (StatusCode::BAD_REQUEST, "Link already exists"),
            Self::LinkNotFoundId(_) | Self::LinkNotFoundSlug(_) => {
                (StatusCode::NOT_FOUND, "Link not found")
            }
            Self::PasteNotFound(_) => (StatusCode::NOT_FOUND, "Paste not found"),
            Self::ImpossibleAffectedRows(_) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "Database returned an impossible number of affected rows",
            ),
            Self::UlidConflict(_) => (StatusCode::INTERNAL_SERVER_ERROR, "ULID conflict real???"),
            Self::Database(_) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "A database error has occured",
            ),
            Self::Io(_) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "An I/O error has occured",
            ),
            Self::Other(_) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "An unknown error has occured",
            ),
        }
        .into_response()
    }
}
